# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/09/22
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import unittest
import os.path
import xml.etree.ElementTree as ET

import pxml
import six
import pkg_resources
from aadict import aadict

import asset

#------------------------------------------------------------------------------
def isEgg(pkg):
  dist = pkg_resources.get_distribution(pkg)
  return os.path.isfile(dist.location)

#------------------------------------------------------------------------------
class TestAsset(unittest.TestCase, pxml.XmlTestMixin):

  maxDiff = None

  #----------------------------------------------------------------------------
  def test_version(self):
    self.assertRegexpMatches(asset.version('pxml'), r'^\d+.\d+.\d+$')

  #----------------------------------------------------------------------------
  def test_load_multi(self):
    self.assertEqual(len(asset.load('asset:test/data/file1.nl')), 1)
    self.assertEqual(
      [str(ast) for ast in asset.load('asset:test/data/file1.nl')],
      ['asset:test/data/file1.nl'])
    self.assertEqual(
      [str(ast) for ast in asset.load('asset:test/data/**.nl')],
      ['asset:test/data/file1.nl',
       'asset:test/data/file2.nl',
       'asset:test/data/subdir/subfile1.nl'])
    self.assertEqual(
      [ast.read() for ast in asset.load('asset:test/data/**.nl')],
      [b'line-1\nline-2',
       b'line-3\n',
       b'sub-file-line-1\n'])

  #----------------------------------------------------------------------------
  def test_load_single(self):
    loaded = []
    for item in asset.load('asset:test/data/file1.nl'):
      loaded.append(item)
    self.assertEqual(len(loaded), 1)
    self.assertEqual(loaded[0].package, 'asset')
    self.assertEqual(loaded[0].name, 'test/data/file1.nl')
    with self.assertRaises(asset.NoSuchAsset) as cm:
      asset.load('asset:no-such-file.ext').peek()

  #----------------------------------------------------------------------------
  def test_load_group_read(self):
    self.assertEqual(
      asset.load('asset:test/data/file1.nl').read(), b'line-1\nline-2')
    self.assertEqual(
      asset.load('asset:test/data/file2.nl').read(), b'line-3\n')
    self.assertEqual(
      asset.load('asset:test/data/*.nl').read(), b'line-1\nline-2line-3\n')
    ag = asset.load('asset:test/data/*.nl')
    self.assertEqual(ag.readline(), b'line-1\n')
    self.assertEqual(ag.readline(), b'line-2')
    self.assertEqual(ag.readline(), b'line-3\n')

  #----------------------------------------------------------------------------
  def test_load_example(self):
    out = ET.Element('nodes')
    for item in asset.load('asset:test/data/**.nl'):
      cur = ET.SubElement(out, 'node', name=item.name)
      cur.text = item.read().decode()
    out = ET.tostring(out)
    chk = b'''\
<nodes>
  <node name="test/data/file1.nl">line-1
line-2</node>
  <node name="test/data/file2.nl">line-3
</node>
  <node name="test/data/subdir/subfile1.nl">sub-file-line-1
</node>
</nodes>
'''
    self.assertXmlEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_listres(self):
    self.assertEqual(
      list(asset.listres('asset', 'test/data', showDirs=True)),
      [
        'test/data/file1.nl',
        'test/data/file2.nl',
        'test/data/subdir/',
        'test/data/subdir/subfile1.nl',
      ])

  #----------------------------------------------------------------------------
  def test_filename_egg(self):
    # NOTE: this requires that `pxml` be installed as a zipped egg, i.e.:
    #   easy_install --zip-ok pxml
    if not isEgg('pxml'):
      raise unittest.SkipTest('package "pxml" must be installed as a zipped egg')
    for item in asset.load('pxml:__init__.py'):
      self.assertIsNone(item.filename)

  #----------------------------------------------------------------------------
  def test_filename_noegg(self):
    # NOTE: this requires that `globre` be installed as an UNzipped pkg, i.e.:
    #   easy_install --always-unzip globre
    if isEgg('globre'):
      raise unittest.SkipTest('package "globre" must not be installed as a zipped egg')
    for item in asset.load('globre:__init__.py'):
      self.assertIsNotNone(item.filename)

  #----------------------------------------------------------------------------
  def test_readWithSize(self):
    self.assertEqual(
      asset.load('asset:test/data/file**').stream().read(),
      b'line-1\nline-2line-3\n')
    self.assertEqual(
      asset.load('asset:test/data/file**').stream().read(1024),
      b'line-1\nline-2line-3\n')
    stream = asset.load('asset:test/data/file**').stream()
    self.assertEqual(stream.read(5), b'line-')
    self.assertEqual(stream.read(5), b'1\nlin')
    self.assertEqual(stream.read(5), b'e-2li')
    self.assertEqual(stream.read(3), b'ne-')
    self.assertEqual(stream.read(3), b'3\n')
    self.assertEqual(stream.read(3), b'')

  #----------------------------------------------------------------------------
  def test_streamIteration(self):
    stream = asset.load('asset:test/data/file**').stream()
    self.assertEqual(stream.readline(), b'line-1\n')
    self.assertEqual(stream.readline(), b'line-2')
    self.assertEqual(stream.readline(), b'line-3\n')
    self.assertEqual(stream.readline(), b'')
    stream = asset.load('asset:test/data/file**').stream()
    chk = list(reversed([
      b'line-1\n',
      b'line-2',
      b'line-3\n',
    ]))
    for line in stream:
      self.assertEqual(line, chk.pop())

  #----------------------------------------------------------------------------
  def test_csv(self):
    import csv
    lines  = [line.decode() for line in asset.load('asset:test/data.csv').stream()]
    reader = csv.reader(lines)
    self.assertEqual(six.next(reader), ['a', 'b', 'c'])
    self.assertEqual(six.next(reader), ['1', '2', '3'])
    with self.assertRaises(StopIteration):
      six.next(reader)


#------------------------------------------------------------------------------
class TestPlugins(unittest.TestCase):

  maxDiff = None

  @staticmethod
  def method(): pass

  #----------------------------------------------------------------------------
  def test_plugin_sorting_intra(self):
    from .plugin import _sort_plugins
    self.assertEqual(
      list(_sort_plugins('myext', [
        aadict(name='foo', after=None, before=None, order=8, replace=False, final=True),
        aadict(name='foo', after=None, before=None, order=2, replace=False, final=False),
        aadict(name='foo', after=None, before=None, order=9, replace=False, final=False),
        aadict(name='foo', after=None, before=None, order=5, replace=True,  final=False),
      ])), [
        aadict(name='foo', after=None, before=None, order=5, replace=True,  final=False),
        aadict(name='foo', after=None, before=None, order=8, replace=False, final=True),
      ])

  #----------------------------------------------------------------------------
  def test_plugin_sorting_inter(self):
    from .plugin import _sort_plugins
    self.assertEqual(
      list(_sort_plugins('myext', [
        aadict(name='a', after=None,  before=None, order=8, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=9, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
      ])), [
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
      ])

  #----------------------------------------------------------------------------
  def test_plugin_sorting_spec_valid(self):
    from .plugin import _sort_plugins
    self.assertEqual(
      list(_sort_plugins('myext', [
        aadict(name='a', after=None,  before=None, order=8, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=9, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
      ], 'c,b,a')), [
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
      ])
    self.assertEqual(
      list(_sort_plugins('myext', [
        aadict(name='a', after=None,  before=None, order=8, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=9, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
      ], '-c')), [
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
      ])

  #----------------------------------------------------------------------------
  def test_plugin_sorting_spec_invalid(self):
    from .plugin import _sort_plugins
    with self.assertRaises(TypeError) as cm:
      list(_sort_plugins('myext', [
        aadict(name='a', after=None,  before=None, order=8, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=9, replace=False, final=False),
        aadict(name='b', after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a', after='b',   before=None, order=5, replace=False, final=True),
        aadict(name='c', after=None,  before='a',  order=0, replace=False, final=False),
      ], '-c,-b'))
    self.assertEqual(
      str(cm.exception),
      "myext plugin type 'a' specified unavailable 'after' dependency 'b'")

  #----------------------------------------------------------------------------
  def test_plugin_sorting_unavailable(self):
    from .plugin import _sort_plugins
    with self.assertRaises(TypeError) as cm:
      list(_sort_plugins('myext', [
        aadict(name='bar', after='foo', before=None, order=0, replace=False, final=False),
      ]))
    self.assertEqual(
      str(cm.exception),
      "myext plugin type 'bar' specified unavailable 'after' dependency 'foo'")
    with self.assertRaises(TypeError) as cm:
      list(_sort_plugins('myext', [
        aadict(name='bar', after=None, before='foo', order=0, replace=False, final=False),
      ]))
    self.assertEqual(
      str(cm.exception),
      "myext plugin type 'bar' specified unavailable 'before' dependency 'foo'")
    with self.assertRaises(TypeError) as cm:
      list(_sort_plugins('myext', [
        aadict(name='foo', after=None, before='bar', order=0, replace=False, final=False),
        aadict(name='bar', after=None, before='foo', order=0, replace=False, final=False),
      ]))
    self.assertEqual(
      str(cm.exception),
      "myext has cyclical dependencies in plugins ['bar', 'foo']")

  #----------------------------------------------------------------------------
  def test_plugin_spec_valid(self):
    import re
    from .plugin import _parse_spec
    self.assertEqual(_parse_spec(None), ())
    self.assertEqual(_parse_spec('*'), ())
    self.assertEqual(_parse_spec('-foo,+bar'), (('-', 'foo'), ('+', 'bar')))
    self.assertEqual(_parse_spec('-foo,-bar'), (('-', 'foo'), ('-', 'bar')))
    self.assertEqual(_parse_spec(' - foo, - bar'), (('-', 'foo'), ('-', 'bar')))
    self.assertEqual(_parse_spec(' - foo - bar '), (('-', 'foo'), ('-', 'bar')))
    self.assertEqual(_parse_spec('foo,?bar'), ((' ', 'foo'), ('?', 'bar')))
    self.assertEqual(_parse_spec('foo,?bar'), ((' ', 'foo'), ('?', 'bar')))
    self.assertEqual(
      _parse_spec('/(foo|bar)/'),
      (('/', re.compile('(foo|bar)')),))
    self.assertEqual(
      _parse_spec('+zig,/(foo|bar)/'),
      (('+', 'zig'), ('/', re.compile('(foo|bar)'))))
    self.assertEqual(
      _parse_spec('/zig,/(foo|bar)/'),
      (('/', re.compile('zig,/(foo|bar)')),))

  #----------------------------------------------------------------------------
  def test_plugin_spec_invalid(self):
    from .plugin import _parse_spec
    with self.assertRaises(ValueError) as cm:
      _parse_spec('foo,-bar')
    self.assertEqual(
      str(cm.exception),
      "invalid mixing of relative (['-']) and absolute ([' ']) prefixes in plugin specification")
    with self.assertRaises(ValueError) as cm:
      _parse_spec('-foo,bar')
    self.assertEqual(
      str(cm.exception),
      "invalid mixing of relative (['-']) and absolute ([' ']) prefixes in plugin specification")
    with self.assertRaises(ValueError) as cm:
      _parse_spec('-foo,?bar')
    self.assertEqual(
      str(cm.exception),
      "invalid mixing of relative (['-']) and absolute (['?']) prefixes in plugin specification")
    with self.assertRaises(ValueError) as cm:
      _parse_spec('/foo')
    self.assertEqual(
      str(cm.exception),
      'regex plugin loading expression must start and end with "/"')

  #----------------------------------------------------------------------------
  def test_plugin_spec_match(self):
    from .plugin import _parse_spec, _match_spec
    self.assertTrue(_match_spec(_parse_spec('*'), 'foo'))
    self.assertTrue(_match_spec(_parse_spec('/foo/'), 'foo'))
    self.assertFalse(_match_spec(_parse_spec('/foo/'), 'bar'))
    self.assertTrue(_match_spec(_parse_spec('/(foo|bar)/'), 'foo'))
    self.assertTrue(_match_spec(_parse_spec('/(foo|bar)/'), 'bar'))
    self.assertTrue(_match_spec(_parse_spec('foo,?bar'), 'foo'))
    self.assertTrue(_match_spec(_parse_spec('foo,?bar'), 'bar'))
    self.assertTrue(_match_spec(_parse_spec('-foo'), 'bar'))
    self.assertFalse(_match_spec(_parse_spec('-foo'), 'foo'))
    self.assertFalse(_match_spec(_parse_spec('foo,?bar'), 'zog'))
    self.assertTrue(_match_spec(_parse_spec('+foo'), 'foo'))
    self.assertTrue(_match_spec(_parse_spec('+foo'), 'bar'))

  #----------------------------------------------------------------------------
  def test_plugin_asset_load(self):
    from .plugin import plugins, _load_asset_plugin
    self.assertEqual(
      _load_asset_plugin('asset.test.TestPlugins.method').handle,
      self.method)
    plugs = list(plugins('test-group', 'asset.test.TestPlugins.method'))
    self.assertEqual(
      [plin.handle for plin in plugs],
      [self.method])
    with self.assertRaises(ValueError) as cm:
      list(plugins('test-group', 'asset.test.TestPlugins.no_such_method'))
    self.assertEqual(
      str(cm.exception),
      "could not load plugin 'asset.test.TestPlugins.no_such_method'")

  #----------------------------------------------------------------------------
  def test_plugin_asset_mixed(self):
    from .plugin import _sort_plugins
    self.assertEqual(
      list(_sort_plugins('myext', [
        aadict(name='a',   after=None,  before=None, order=8, replace=False, final=False),
        aadict(name='b',   after=None,  before=None, order=9, replace=False, final=False),
        aadict(name='b',   after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='a',   after='b',   before=None, order=5, replace=False, final=True),
        aadict(name='c',   after=None,  before='b',  order=0, replace=False, final=False),
        aadict(name='d.e', after='b',   before='a',  order=0, replace=False, final=False),
      ], '+d.e')), [
        aadict(name='c',   after=None,  before='b',  order=0, replace=False, final=False),
        aadict(name='b',   after=None,  before=None, order=2, replace=False, final=True),
        aadict(name='d.e', after='b', before='a',  order=0, replace=False, final=False),
        aadict(name='a',   after='b',   before=None, order=5, replace=False, final=True),
      ])

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2015/11/12
# copy: (C) Copyright 2015-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import re
import logging

import six
import pkg_resources
from aadict import aadict
import asset

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
def plugins(group, spec=None):
  # TODO: share this documentation with `../doc/plugin.rst`...
  '''
  Returns a list of plugins for the specified setuptools-style
  entrypoint `group`. This is just a wrapper around
  `pkg_resources.iter_entry_points` that allows the plugins to sort
  and override themselves. The plugins have the following attributes
  of note:

  * `name` : the name of the plugin type

  * `handle` : the actual plugin symbol

  * `entrypoint` : for registered plugins, the setuptools EntryPoint
    object, otherwise ``None``

  The optional `spec` parameter controls how and what plugins are
  loaded. If it is ``None`` or the special value ``'*'``, then the
  normal plugin loading will occur, i.e. all registered plugins will
  be loaded and their self-declared ordering and dependencies will be
  applied.

  Otherwise, the `spec` is taken as a comma- or whitespace-separated
  list of plugins to load. In this mode, the `spec` can either specify
  an exact list of plugins to load, in the specified order, referred
  to as an "absolute" spec. Otherwise, it is a "relative" spec, which
  indicates that it only adjusts the standard registered plugin
  loading. A spec is a list of either absolute or relative
  instructions, and they cannot be mixed.

  In either mode, a plugin is identified either by name for registered
  plugins (e.g. ``foo``), or by fully-qualified Python module and
  symbol name for unregistered plugins
  (e.g. ``package.module.symbol``).

  Plugins in an absolute spec are loaded in the order specified and
  can be optionally prefixed with the following special characters:

  * ``'?'`` : the specified plugin should be loaded if available.  If
    it is not registered, cannot be found, or cannot be loaded, then
    it is ignored (a DEBUG log message will be emitted, however).

  Plugins in a relative spec are always prefixed with at least one of
  the following special characters:

  * ``'-'`` : removes the specified plugin; this does not affect
    plugin ordering, it only removes the plugin from the loaded
    list. If the plugin does not exist, no error is thrown.

  * ``'+'`` : adds or requires the specified plugin to the loaded
    set. If the plugin is not a named/registered plugin, then it will
    be loaded as an asset-symbol, i.e. a Python-dotted module and
    symbol name. If the plugin does not exist or cannot be loaded,
    this will throw an error. It does not affect the plugin ordering
    of registered plugins.

  * ``'/'`` : the plugin name is taken as a regular expression that
    will be used to match plugin names and it must terminate in a
    slash. Note that this must be the **last** element in the spec
    list.

  Examples:

  * ``'*'`` : load all registered plugins.

  * ``'foo,bar'`` : load the "foo" plugin, then the "bar" plugin.

  * ``'foo,?bar'`` : load the "foo" plugin and if the "bar" plugin
    exists, load it too.

  * ``'-zig'`` : load all registered plugins except the "zig" plugin.

  * ``'+pkg.foo.bar'`` : load all registered plugins and then load
    the "pkg.foo.bar" Python symbol.

  * ``'pkg.foo.bar'`` : load only the "pkg.foo.bar" Python symbol.
  '''
  spec   = _parse_spec(spec)
  plugs  = list(_get_registered_plugins(group, spec))
  plugs  += list(_get_unregistered_plugins(group, plugs, spec))
  return _sort_plugins(group, plugs, spec)

#------------------------------------------------------------------------------
# relative specs:
SPEC_ADD       = '+'
SPEC_REM       = '-'
SPEC_RE        = '/'
# absolute specs:
SPEC_SET       = ' '
SPEC_OPT       = '?'
SPEC_ALL       = '*'
_specsep_cre   = re.compile(r'(^|\s+)([-+?])\s*')
_specname_cre  = re.compile(r'^[a-zA-Z_]')
_spec_rel      = set((SPEC_ADD, SPEC_REM, SPEC_RE))
_spec_abs      = set((SPEC_SET, SPEC_OPT))

#------------------------------------------------------------------------------
def _parse_spec(spec):
  if isinstance(spec, tuple):
    return spec
  wspec = spec.strip() if spec else None
  if not wspec or wspec == SPEC_ALL:
    return ()
  respec = None
  if wspec.endswith('/'):
    if len(wspec.split('/')) <= 2:
      raise ValueError(
        'regex plugin loading expression must start and end with "/"')
    wspec, respec = wspec.split('/', 1)
    respec = respec[:-1]
    if wspec.strip() \
        and not ( wspec.strip().endswith(',') or wspec != wspec.rstrip() ):
      raise ValueError(
        'regex plugin loading expression must start and end with "/"')
  wspec = wspec.replace(',', ' ')
  wspec = _specsep_cre.sub(r' \2', wspec)
  wspec = [e.strip() for e in wspec.split()]
  ret = []
  for item in wspec:
    if not item.strip():
      continue
    item = (SPEC_SET, item)
    if item[1] and item[1][0] in (SPEC_ADD, SPEC_REM, SPEC_OPT):
      item = (item[1][0], item[1][1:].strip())
    if item[1] and item[1][0] in (SPEC_RE,):
      raise ValueError(
        'regex plugin loading expression must start and end with "/"')
    if not _specname_cre.match(item[1]):
      raise ValueError(
        'invalid plugin name in specification expression: %r' % (item[1],))
    ret.append(item)
  if respec:
    ret.append((SPEC_RE, re.compile(respec)))
  ops = set([item[0] for item in ret])
  rel = ops & _spec_rel
  abs = ops & _spec_abs
  if bool(ops & rel) and bool(ops & abs):
    raise ValueError(
      'invalid mixing of relative (%r) and absolute (%r) prefixes in plugin specification'
      % (list(rel), list(abs)))
  return tuple(ret)

#------------------------------------------------------------------------------
def _match_spec(spec, name):
  if not spec:
    return True
  for idx, item in enumerate(spec):
    if item[0] == SPEC_RE:
      if item[1].match(name):
        return True
      if ( idx + 1 ) >= len(spec):
        return False
      continue
    if item[1] != name:
      continue
    if item[0] == SPEC_REM:
      return False
    return True
  if spec[0][0] in _spec_rel:
    return True
  return False

#------------------------------------------------------------------------------
def _decorate_plugin(plugin):
  plugin.after   = getattr(plugin.handle, 'after',   None)
  plugin.before  = getattr(plugin.handle, 'before',  None)
  plugin.order   = getattr(plugin.handle, 'order',   0)
  plugin.replace = getattr(plugin.handle, 'replace', False)
  plugin.final   = getattr(plugin.handle, 'final',   False)

#------------------------------------------------------------------------------
def _get_registered_plugins(group, spec=None):
  spec = _parse_spec(spec)
  for entrypoint in pkg_resources.iter_entry_points(group):
    plugin = aadict(
      name         = entrypoint.name,
      entrypoint   = entrypoint,
    )
    if _match_spec(spec, plugin.name):
      plugin.handle  = entrypoint.load()
      _decorate_plugin(plugin)
      yield plugin

#------------------------------------------------------------------------------
def _load_asset_plugin(spec):
  plugin = aadict(
    name         = spec,
    entrypoint   = None,
    handle       = asset.symbol(spec),
  )
  _decorate_plugin(plugin)
  return plugin

#------------------------------------------------------------------------------
def _get_unregistered_plugins(group, plugins, spec=None):
  spec = _parse_spec(spec)
  names = [plug.name for plug in plugins]
  for item in spec:
    if item[0] in (SPEC_REM, SPEC_RE) or item[1] in names:
      continue
    try:
      yield _load_asset_plugin(item[1])
    except Exception:
      if item[0] in (SPEC_OPT,):
        log.debug('could not load optional plugin %r', item[1])
        continue
      raise ValueError('could not load plugin %r' % (item[1],))

#------------------------------------------------------------------------------
def _sort_plugins(group, plugins, spec=None):
  spec = _parse_spec(spec)
  lut = dict()
  for plugin in plugins:
    if plugin.name not in lut:
      lut[plugin.name] = []
    lut[plugin.name].append(plugin)
  # order the plugins within each named group by order/replace/final
  for name, plugs in list(lut.items()):
    plugs = sorted(plugs, key=lambda plug: plug.order)
    for idx, plug in enumerate(plugs):
      if plug.final:
        plugs = plugs[:idx + 1]
        break
    for idx, plug in enumerate(reversed(plugs)):
      if plug.replace:
        plugs = plugs[len(plugs) - idx - 1:]
    lut[name] = plugs
  # apply `spec`
  names = [name for name in sorted(lut.keys()) if _match_spec(spec, name)]
  if spec and spec[0][0] not in _spec_rel:
    snames = []
    for item in spec:
      if item[1] in names:
        snames.append(item[1])
        continue
      if item[0] == SPEC_OPT:
        continue
      raise TypeError(
        '%s plugin spec %r specified unavailable dependency %r'
        % (group, spec, item[1]))
    for name in snames:
      for plug in lut[name]:
        yield plug
    return
  # order the named groups by after/before
  reqs = dict()
  for name in names:
    reqs[name] = aadict(after=[], before=[])
    for plug in lut[name]:
      if plug.after:
        reqs[name].after += [p.strip() for p in plug.after.split(',')]
      if plug.before:
        reqs[name].before += [p.strip() for p in plug.before.split(',')]
    reqs[name].after  = [aft for aft in reqs[name].after if aft]
    reqs[name].before = [bef for bef in reqs[name].before if bef]
  # simplify after/before evaluation by pushing before's into after's
  for name in names:
    for aft in reqs[name].after:
      opt = aft.startswith(SPEC_OPT)
      if not opt and aft not in names:
        raise TypeError(
          '%s plugin type %r specified unavailable \'after\' dependency %r'
          % (group, name, aft))
    for bef in reqs[name].before:
      opt = bef.startswith(SPEC_OPT)
      if opt:
        bef = bef[1:]
      if bef not in names:
        if not opt:
          raise TypeError(
            '%s plugin type %r specified unavailable \'before\' dependency %r'
            % (group, name, bef))
        continue
      reqs[bef].after.append(name)
  # created the ordered list obeying the 'after' rules
  snames = []
  while len(snames) < len(names):
    cur = len(snames)
    for name in names:
      if name in snames:
        continue
      ok = True
      for aft in reqs[name].after:
        opt = aft.startswith(SPEC_OPT)
        if opt:
          aft = aft[1:]
        if aft not in snames and aft in names:
          ok = False
      if ok:
        snames.append(name)
    if len(snames) == cur:
      raise TypeError(
        '%s has cyclical dependencies in plugins %r'
        % (group, sorted(list(set(names) - set(snames))),))
  for name in snames:
    for plug in lut[name]:
      yield plug

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------
